'use strict';

const deepFreeze = require('deep-freeze-strict');
const Promise = require('bluebird');
const DEFAULTS = require('./constants').DEFAULTS;
const ERROR = require('./constants').ERROR;
const AppErr = require('punch-error');

const ProxyManager = require('punch-proxy-manager');

const scraper = require('./scraper/');


class ScraperManager {
  /**
   *
   */
  constructor(config) {
    this.config = config = config || {};
    this.proxyManager = new ProxyManager(config.proxyManagerConfig || {});
    this.maxTry = config.maxTry || DEFAULTS.LINK_MAX_TRY;

    this.proxyManager.start();
  }
  
  start(){
    // Start loading proxies.
    return this.proxyManager.start();
  }

  stop(){
    return this.proxyManager.stop();
  }


  scrapeLink(link, config, opt_count){
    let onUpdate = config.onUpdate;
    return this.proxyManager.ensureLoaded()
    .then(() => {
      opt_count = opt_count || 1;
      let conf = Object.assign({}, this.config.scraperConfig || {}, config || {});

      let proxy = this.proxyManager.get();
      console.log('Scraping:', link, 'using:', proxy.getIP());
      conf.proxy = proxy;
      proxy.markTried(link);

      return scraper.scrape(link, conf)
      .then((html) => {
        proxy.markSucceeded(link);
        if (onUpdate) onUpdate({link, successful: true, proxy: proxy.getIP()});
        return {html};
      })
      .catch((error) =>  {
        // console.log('ScraperManager received error. Marking proxy failed.');
        proxy.markFailed(link);
        if (onUpdate) onUpdate({link, successful: false, proxy: proxy.getIP()});
        if (opt_count <= this.maxTry) return this.scrapeLink(link, config, ++opt_count);
        return AppErr.reject(error, ERROR.LINK_MAX_TRY_ERROR);
      });
    });
  }

  /**
   *
   */
  scrape(links, config) {
    config = Object.assign({}, this.config.scraperConfig || {}, config || {});
    let status = {links: {}, successful: [], failed: [], remaining: links.length};

    let onUpdate = config.onUpdate;
    
    // override config onUpdate.
    config.onUpdate = (update) => {
      status.links[update.link].attempts++;
      if (onUpdate) onUpdate(status, update);
    };

    let scraperResults = links.map((link) => {
      status.links[link] = {result: null, attempts: 0};
      return this.scrapeLink(link, config)
      .tap((response) => {
        status.successful.push(link);
        status.links[link].result = response.html;
      })
      .catch(err => {
        status.failed.push(link);
        // Allow promise to resolve.
      })
      .finally(() => {
        status.remaining--;
      });
    });

    return Promise.all(scraperResults)
    .tap(() => onUpdate && onUpdate(status));
  }
}

ScraperManager.STRATEGIES = require('./scraper/strategies/constants').STRATEGY;

module.exports = ScraperManager;